/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2002 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

#include "eel.h"
#include "proc.h"
#include "epsnet.h"
#include "grep.h"
#include "html.h"
#include "tags.h"
#include "c.h"
#include "dired.h"
#include "colcode.h"
#include "vbasic.h"
#include "kill.h"
#include "bufed.h"

// For updating from 11.00 or later.
// If we're updating from a beta version of 11.00, first load
// an updater to bring us to 11.00.

when_loading()
{
	if (!strncmp(eel_version, "11.00b", 6)) {
		save_var bufnum = zap(HELPBUF);
		bprintf("%s is from a beta version of Epsilon.\n"
				"The above state file is for Epsilon version %s.\n"
				"This Epsilon executable is from Epsilon version %s.\n\n"
				"I must first load a separate updater to bring this\n"
				"state file to 11.00.  Press Enter to continue.\n",
				state_file, eel_version, EEL_MINOR_VERSION);
		view_help(" Beta Version Detected ", 0);
		update_autoload("find-file", "oquickup");
	}
}

virtual_mark_column()
{
	return (mark_to_column >= 0) ? mark_to_column : get_column(mark);
}

do_shift_selects()
{
	if (!shift_selects || key < 512
		  || (NUMPLAIN(key) < KEYEND || NUMPLAIN(key) > KEYPGUP)
		  && (NUMPLAIN(key) < NUMDIGIT(0) || NUMPLAIN(key) > NUMDIGIT(9))
	   ) {	// only affects cursor keys or numpad keys
		if (shift_selecting == 2)		// except at prompts
			highlight_off();
		return;
	}
	if (was_key_shifted()) {
		if (!is_highlight_on())		// if shift key is down, begin/extend
			do_set_mark(point);		// selection
		set_region_type(REGNORM);
		highlight_on();
		shift_selecting = 1;
	} else if (shift_selecting)	// if shift was just removed, remove highlight
		highlight_off();
}

#define BROWSER_BUF "Symbols"

#define BSC_DLL "bsc.dll" /* Compiler-specific DLL file from Microsoft. */

// This takes the address of a pointer, and makes the pointer point
// to a BSC handle.
#define BSCOpen(browsefile, ppbsc) \
								   ((int) call_dll(BSC_DLL, \
								   "BSCOpen", "c", "sr", browsefile, ppbsc))

// This takes a pointer to a BSC handle.
#define BSCClose(pbsc) ((int) call_dll(BSC_DLL, "BSCClose", "c", "p", pbsc))

#define BSCGetOverloadArray(pbsc, name, mbf, ppinst, pnum) \
														   ((int) call_dll(BSC_DLL, \
														   "BSCGetOverloadArray", "c", "psirr", \
														   pbsc, name, mbf, ppinst, pnum))


#define BSCIinstInfo(pbsc, inst, name, typ, atr) \
												 ((int) call_dll(BSC_DLL, \
												 "BSCIinstInfo", "c", "pirpp", \
												 pbsc, inst, name, typ, atr))

#define BSCFormatDname(pbsc, name) \
								   ((int) call_dll(BSC_DLL, \
								   "BSCFormatDname", "c", "pp", \
								   pbsc, name))

#define BSCGetDefArray(pbsc, inst, ppidef, pnum) \
												 ((int) call_dll(BSC_DLL, \
												 "BSCGetDefArray", "c", "pirr", \
												 pbsc, inst, ppidef, pnum))

#define BSCGetRefArray(pbsc, inst, ppidef, pnum) \
												 ((int) call_dll(BSC_DLL, \
												 "BSCGetRefArray", "c", "pirr", \
												 pbsc, inst, ppidef, pnum))


#define BSCIdefInfo(pbsc, idef, pmodname, plineno) \
												   ((int) call_dll(BSC_DLL, \
												   "BSCIdefInfo", "c", "pirp", \
												   pbsc, idef, pmodname, plineno))													  

#define BSCSzFrTyp(pbsc, typ) \
							  ((int) call_dll(BSC_DLL, \
							  "BSCSzFrTyp", "c", "pi", \
							  pbsc, typ))

#define BSCSzFrAtr(pbsc, atr) \
							  ((int) call_dll(BSC_DLL, \
							  "BSCSzFrAtr", "c", "pi", \
							  pbsc, atr))

#define BSCDisposeArray(pbsc, any_array) \
										 ((int) call_dll(BSC_DLL, \
										 "BSCDisposeArray", "c", "pp", \
										 pbsc, any_array))

void get_inst_data(int *pbsc, int inst, char *sym_name, char *sym_descr)
{
	char sym_type[512];
	char sym_attr[512];
	char *undecor_name;
	char typ[1];
	short atr[1];

	BSCIinstInfo(pbsc, inst, &undecor_name, typ, atr);

	strcpy(sym_name, (char *) BSCFormatDname(pbsc, undecor_name));
	strcpy(sym_type, (char *) BSCSzFrTyp(pbsc, *typ));
	strcpy(sym_attr, (char *) BSCSzFrAtr(pbsc, *atr));

	sprintf(sym_descr, "%s%s%s",
			sym_attr[0] ? sym_attr : "",
			sym_attr[0] ? " " : "",
			sym_type
		   );
}

int dump_defs(int *pbsc, char *bsc_filename, char *symbol, int *insts, int num_insts)
{
	char sym_name[512];
	char sym_descr[512];
	int r = 0;
	int i, j;

	bufnum = zap(BROWSER_BUF);

	use_black(); bprintf("Browser File: ");
	use_blue();  bprintf("%s\n", bsc_filename);

	use_black(); bprintf("Symbol: ");
	use_blue();  bprintf("%s\n", symbol);

	use_black();bprintf("\n%s:\n", "Definitions");

	for (j = 0; j < num_insts; j++){
		int *defs, num_defs;
		int inst = insts[j];

		get_inst_data(pbsc, inst, sym_name, sym_descr);
		BSCGetDefArray(pbsc, inst, &defs, &num_defs);
		for (i = 0; i < num_defs; i++){
			char *modname;
			short line_num[1];
			BSCIdefInfo(pbsc, defs[i], &modname, line_num);
			r++;

			// The file name.
			use_black();
			bprintf("    %s(%d): ", modname, *line_num);

			// The symbol name.
			use_blue();
			bprintf("%s", sym_name);

			// The symbol description.
			use_green();
			bprintf(" (%s)", sym_descr);
			bprintf("\n");
			use_black(); // flush
		}
	}
	point = 0;
	return r;
}

int bufed_when_displaying();

do_bufed(flags)
{
	char find[FNAMELEN], msg[FNAMELEN];
	int count;

	if (in_bufed()) {
		bufnum = window_bufnum;
		set_read_only(0);
		if (bufed_getname(find)) {
			quote_bufname(find);
			strcat(find, " ");
		} else
			*find = 0;
		bufed_restore_orig();
		make_buffer_list(flags, BUFED_BUF);
		window_handle = _bufed_window;
		window_bufnum = name_to_bufnum(BUFED_BUF);
		point = 0;
	} else {
		_bufed_orig_window = window_handle;
		_bufed_orig_buffer = window_bufnum;
		_bufed_orig_highlight =_highlight_control;
		strcpy(find, bufname);
		quote_bufname(find);
		strcat(find, " ");
		make_buffer_list(flags, BUFED_BUF);
		build_first = 1;
		_bufed_window = add_popup(2, 2, bufed_width, screen_lines - 6,
								  ALL_BORD(BSINGLE), name_to_bufnum(BUFED_BUF));
		if (is_gui && want_gui_prompts) {
			count = lines_between(0, size(), 0) + 1;
			mouse_cursor = &left_pointer;
			_bufed_caption_window = add_popup(0, 0, bufed_width, 1, 0,
											  create(BUFED_CAPTION_BUF));
			buf_zap(bufnum);
			bprintf("    Size     Buffer             File ");
			point = 0;
			system_window = 1;
			undo_size = 0;
			mouse_cursor = &left_pointer;
			window_handle = _bufed_window;

			if (flags & BUFED_EXIT) {
				sprintf(msg, "%d buffer%s not saved.  Exit?", count,
						(count == 1) ? "" : "s");
				display_dialog_box("AskExitBox", msg,
								   _bufed_caption_window, _bufed_window, 0,
								   "E&xit", "&Cancel", "");
			} else if (flags & BUFED_SAVE) {
				sprintf(msg, "Save %d unsaved buffers first? "
						"(! to never ask)", count);
				display_dialog_box("AskSaveBox", msg,
								   _bufed_caption_window, _bufed_window, 0,
								   "&Save", "&No Save", "&Cancel");
			} else if (flags & BUFED_SWITCH)
				display_dialog_box("SwitchBufBox", "Select a Buffer",
								   _bufed_caption_window, _bufed_window, 0,
								   "OK", "Cancel", "Help");
			else
				display_dialog_box("CaptionBox", "Select a Buffer",
								   _bufed_caption_window, _bufed_window, 0,
								   "OK", "Cancel", "Help");
		}
		window_handle = _bufed_window;
	}
	set_menu_look(_bufed_window);
	if (flags & (BUFED_EXIT | BUFED_SAVE))
		remove_line_highlight();
	else
		make_line_highlight();
	window_title(_bufed_window, BTOP, TITLELEFT(4),
				 " Size     Buffer             File ");
	undo_size = 0;
	display_column = 0;	/* horizontally scroll this window */
	popup_near_window(_bufed_window, _bufed_orig_window);
	bufed_mode();
	if (*find)
		col_search(find, BNAME_COL);
	to_bufed_column();
	modified = 0;
	goal_column = BNAME_COL;
	when_displaying = bufed_when_displaying;
}

newest_before(t)
{
	int i, newest_time = 0, newest = 0;

	save_var bufnum;
	i = buf_list(0, 2);	/* start with most recently created buf */
	do {
		bufnum = i;
		if (buf_accessed >= t || *bufname == '-'
			  || !strcmp(bufname, MSG_BUFFER) && !buf_accessed)
			continue;
		if (!newest || buf_accessed > newest_time) {
			newest_time = buf_accessed;
			newest = i;
		}
	} while (i = buf_list(-1, 1));
	return newest;
}

give_prev_buf()		/* return buf # of previous buffer, or 0 if none */
{
	int i, j;

	if (_prev_buffer != bufnum && buf_exist(_prev_buffer))
		return _prev_buffer;
	for (i = _prev_num; --i >= 0; ) {	/* start with oldest at end */
		j = _prev_old_buffers[i];
		if (buf_exist(j) && j != bufnum && buf_in_window(j) < 0
			  && strcmp(bufnum_to_name(j), MSG_BUFFER))
			return j;
	}
	for (i = buf_accessed_clock; j = newest_before(i);
		 i = get_buf_accessed(j))
		if (j != bufnum)
			return j;
	return 0;
}

buf_xfer_colors(num, from, to)
{
	int orig = bufnum, pos, color, colstart, colend;

	if (to < from)
		{ pos = from; from = to; to = pos; }
	save_var narrow_start = 0, narrow_end = 0;
	buf_xfer(num, from, to);
	for (pos = from; pos < to;) {
		color = get_character_color(pos, &colstart, &colend);
		bufnum = num;
		if (colstart < from)	// Only change the part we just copied.
			colstart = from;
		if (colend > to)
			colend = to;
		set_character_color(mark + colstart - from,
							mark + colend - from, color);
		bufnum = orig;
		pos = colend;
	}
}

indent_case(orig)	// indent case-like statement at orig
{
	point = orig;
	for (;;) {		// look for case-like statement, or {}'s
		cind_search("([^a-zA-Z0-9_]!(case|default|public|"
					"private|protected)[^a-zA-Z0-9_])|[{}]");
		if (in_c_comment(c_look_back))
			continue;
		else if (curchar() != '}')
			break;
		point++;
		if (!c_rev_skip_level('{', '}')) // move before matching {
			longjmp(&c_indent_giveup, 1);
	}
	if (curchar() != '{')	// copy indentation from prior case
		return get_indentation(point);

	cind_space();	// skip over switch-like construct before {
	if (character(point - 1) == ')') // back past () of switch
		if (!c_rev_skip_level('(', ')'))
			longjmp(&c_indent_giveup, 1);
	to_indentation();
	past_labels();	// might be "case 3: switch ()", get to switch
	point += parse_string(1, "[ \t\f]*");
	return current_column() + get_c_indent() +
			(matches_at(orig, 1, "(public|private|protected)</word>")
			 ? c_access_spec_offset : c_case_offset);
}

int indent_just_open(ind)		// Indent a line with just { on it.
{
	int extra = 0;

	if (!Topindent && !ind && c_statement_start())
		extra = get_c_indent();
	if (ind && !Closeback)
		extra = get_c_indent();
	ind += c_brace_offset + extra;
	if (!ind)
		ind += c_top_braces;
	return ind;
}

int indent_just_open(ind)		// Indent a line with just { on it.
{
	int extra = 0;

	if (!Topindent && !ind && c_statement_start())
		extra = get_c_indent();
	if (ind && !Closeback)
		extra = get_c_indent();
	ind += c_brace_offset + extra;
	if (!ind)
		ind += c_top_braces;
	return ind;
}

indent_after_statement(orig) // orig is place needing indentation, for toplevel
{
	int lastcont, i, c, atparen = -1;

	for (;;) { /* was only the end of a larger statement, find start */
		do {	/* look back for line that ends with ;{}: */
			lastcont = point;
			to_begin_line();
			if (cind_try_space())
				break;
			if ((c = character(point - 1)) == ')') {
				atparen = lastcont;
				if (!c_rev_skip_level('(', ')'))
					break;
				to_begin_line();
			}
		} while (!index(";{}:", c));
		point = lastcont;	/* use indentation of line */
		to_c_indentation();	/* just after it */
		if (parse_string(RE_FORWARD, "(}[ \t\f]*)?e!lse[^a-zA-Z0-9_]")) {
			point = matchend;
			if (!to_matching_if())
				break;
		} else if (parse_string(RE_FORWARD,
								"(}[ \t\f]*)?!(catch|finally)[^a-zA-Z0-9_]")) {
			point = matchend;
			if (!to_matching_try())
				break;
		} else
			break;
	}
	i = current_column();
	if (!i && atparen != -1 && (Topindent || !c_statement_start()))
		i = get_indentation(atparen);	// 2nd func param decl
	if ((atparen = parse_string(1, "for[ \t\f]*%([ \t\f]*"))
		  && for_unclosed(point, orig))
		i = get_column(point + atparen); // part of "for" statement
	if (i < c_top_braces)
		return 0;	// 1st line of new func
	return i;
}

indent_after_block(orig)
{
	int ind;

	c_rev_skip_level('{', '}');	/* find matching { */
	to_begin_line();
	ind = get_indentation(point);
	if (Topindent && (ind == 0 || parse_string(1, "[ \t]*{")
					  && (ind <= c_brace_offset || ind <= c_top_braces)))
		return 0;	// found a func def block
	return indent_after_statement(orig);
}

indent_c_contin(ind, orig, lastl)	/* indent continuation line */
{
	int c, plevel = 0, qlevel = 0, findc;

	if (!c_align_contin_lines)	// never line up
		return c_default_indent(ind, orig, 1);
	to_begin_line();
	save_var narrow_start = point;
	point = orig;
	findc = character(orig);
	if (c_align_open_paren && parse_string(1, "%(([ \t]|//|/%*)*$"))
		return ind;
	while (re_search(-1, "[()?:\"'\n]|[=!<>]?=")) {
		switch (c = curchar()) {
			case '(':	/* indent after unmatched lparen */
				if (plevel-- > 0)
					break;
				if (c_align_open_paren && findc != ')')
					return ind + get_c_indent();
indhere:		if (findc != ')')
					point++;
				if (!parse_string(1, "[ \t\f]*/[*/]"))
					point += parse_string(1, "[ \t\f]*![^ \t\f\n]");
				return limit_c_contin_indent(ind);
			case ')':
				plevel++;
				break;
			case '?':	/* indent : to matching ? */
				if (qlevel-- > 0 || plevel > 0 || findc != ':')
					break;
				point--;
				goto indhere;
			case ':':
				qlevel++;
				break;
			case '=': case '!':
				if (character(point + 1) == '=' || plevel > 0
					  || character(lastl - 1) == ',' || findc == ')')
					break;
				goto indhere;
			case '\n':		 // skip over comments
				cind_try_space(); // at ends of lines
				break;
			case '"': case '\'':
				point++;
				c_skip_item(c, RE_REVERSE);
				break;
		}
	}
	if (character(lastl - 1) == ',') {
		if (character(lastl - 2) != ')') {
			point = lastl;
			to_begin_line();
		}
		point += parse_string(1, "(.*[ \t\f]on)?[ \t\f]+");
		if (in_c_comment(c_look_back))
			to_indentation();
		return limit_c_contin_indent(ind);
	}
	return c_default_indent(ind, orig, 2);
}

check_running_processes()
{
	char *buf, *open = 0;

	if (!process_warn_on_exit)
		return;
	buf = buffer_list(1);
	save_var bufnum;
	do {
		if (is_process_buffer(name_to_bufnum(buf)))
			if (process_warn_on_exit == 2 || !exit_this_process(buf)) {
				open = buf;
				break;
			}
	} while (buf = buffer_list(0));
	restore_vars();
	if (open) {
		locate_window(open, "");
		if (process_warn_on_exit != 2)
			error("Couldn't exit from concurrent process.");
		else if (!ask_yn("Confirm exit", "A process is running, confirm exit? ",
						"Yes", "No", 1))
			error("Exiting canceled.");
	}
}

char _fundamental_mode_name[] = "Fundamental";

do_mode_default_settings(int retain_colors)
{
	indenter = indenter.default;
	if (!indenter)
		indenter = prev_indenter;
	auto_indent = auto_indent.default;
	buffer_maybe_break_line = 0;
	strcpy(comment_start, comment_start.default);
	strcpy(comment_pattern, comment_pattern.default);
	strcpy(comment_begin, comment_begin.default);
	strcpy(comment_end, comment_end.default);
	mode_keys = buf_tab;		/* use these keys */
	mode_move_level = 0;
	major_mode = _fundamental_mode_name;
	auto_show_matching_characters = fundamental_auto_show_delim_chars;
	when_displaying = 0;
	recolor_range = recolor_from_here = 0;
	search_continuation = 0;
	if (!retain_colors) {
		drop_coloring(bufnum);
		coloring_flags &= ~(COLOR_MINIMAL | COLOR_INVALIDATE_FORWARD
							| COLOR_INVALIDATE_BACKWARD | COLOR_INVALIDATE_RESETS
							| COLOR_RETAIN_NARROWING);
	}
	make_mode();
}

dired_sort_by_date()	/* insert stuff in each line, return length */
{
	char line[FNAMELEN], *p, *months;
	int end, hour, i, month;

	months = "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ";
	point = 0;
	i = get_dired_item("dired-sort-by-date-", 1);
	if (i)		// Call style-specific function.
		return (*(int (*)())i)();
	do {
		end = give_end_line();
		if (point == end)
			break;
		grab(point, end, line);
		hour = numtoi(line + 15);
		if (hour == 12 && strchr("ap", line[23]))
			hour = 0;
		if (line[23] == 'p')
			hour += 12;
		line[8] = 0;
		p = strstr(months, line + 5);
		if (!p) {
			say("Unrecognized month %s, can't sort.", line + 5);
			p = months;
		}
		month = 1 + (p - months) / 4;	// Compute month 1..12.

		move_to_column(goal_column);
		bprintf("%.4s%2d%.2s%2d%.5s",
				line + 9, month, line + 2, hour, line + 18);
		point = end;
	} while (nl_forward());
	return 15;
}

substitute_func_name(char *m)
{
	char *p, *s, tmp[FNAMELEN];

	for (p = m; p = strchr(p, '%'); ) {	// look for %f in string.
		if (p[1] == 'f') {
			strcpy(tmp, p + 2);
			if (display_func_name_method() == 1 && *display_func_name) {
				for (s = display_func_name; *s && ptrlen(s) > 1
						 && ptrlen(p) > 2; ) {
					if (*s == '%')
						*p++ = '%';
					*p++ = *s++;
				}
				*p++ = ' ';
				*p = 0;
			}
			if (ptrlen(p) > strlen(tmp) + 2)
				strcpy(p, tmp);
		} else		// Skip past %% sequences.
			p += p[1] ? 2 : 1;
	}
}

ftp_activity(int activity, int buf, int from, int to)
{
	char msg[FNAMELEN];
	int show_log = 0, orig_buf = bufnum;

	if (!buf_exist(buf))
		return;
	save_var iter, has_arg;
	save_var bufnum = buf;
	switch (activity) {
		case NET_DONE:
			if (ftp_job->operation == FTP_SEND)
				show_text(0, 1, "%s written.", filename);
			if (translation_type == FILETYPE_AUTO
				&& ftp_job->operation == FTP_RECV)
				translation_type = new_buffer_translation_type;
			if (!modified && !point) {
				save_var undo_size = 0;
				undo_mainloop();
			}
			free(ftp_job);		// Fall through.
		case NET_LOG_DONE:
			ftp_job = 0;
			set_mode_message(NULL);
			break;
		case NET_RECV:
			if (!from) {		// First block.
				position_in_all_windows(buf, 0);
				if (mode_keys == buf_tab && !*get_extension(filename))
					suffix_none();		// Switch out of Fundamental mode.
			}
			show_ftp_percent("got", from);
			break;
		case NET_SEND:
//			ftp_job->so_far = from;
			ftp_job->file_size = to;
			show_ftp_percent("sent", from);
			break;
		case NET_LOG_WRITE:
			save_var point = from;	// Get file size from msg.
			if (re_search(1, "^150 ") && re_search(1, "[0-9]+ bytes")) {
				char filesize[100];
				grab(matchstart, matchend, filesize);
				if (numtoi(filesize) && !got_bad_number)
					ftp_job->file_size = numtoi(filesize);
			}
			point = from;
			if (found_ftp_status("^220 "))	// Welcome msg.
				set_both_mode_message("[FTP: Sending user name] ");
			if (found_ftp_status("^331 "))	// Host wants password.
				set_both_mode_message("[FTP: Sending password] ");
			if (found_ftp_status("^230 "))	// User logged in.
				set_both_mode_message("[FTP: Working] ");
			if (found_ftp_status("^530 "))	// Bad password.
				discard_password_data();
			if (re_search(1, "^[45][0-9][0-9] ") && !parse_string(-1, "PASV\n... ")) {
				parse_string(1, ".*", msg);		// Some error code.
				say("FTP error: %s", msg);
				// Don't go to log if we're just creating a new file.
				if (ftp_job->operation != FTP_RECV
					|| !parse_string(-1, "^550 .*"))
					show_log = 1;
			}
			break;
	}
	if (buffer_ftp_activity)	// Do some buffer-specific FTP work.
		buffer_ftp_activity(activity, buf, from, to);
	switch (activity) {		// These don't automatically refresh later.
		case NET_DONE:
		case NET_LOG_DONE:
		case NET_SEND:
			refresh();
			break;
	}
	if (show_log) {
		if (num_of_popups() > 0 || orig_buf != ftp_job->buf) {
			say("FTP errors occurred, see buffer %s.",
				buf_exist(ftp_job->log_buf) ?
				bufnum_to_name(ftp_job->log_buf) : "<deleted>");
			if (buf_exist(ftp_job->buf) && ftp_job->operation == FTP_SEND) {
				save_var bufnum = ftp_job->buf;
				modified = 1;
			}
		} else {
			restore_vars();
			save_var iter, has_arg;
			save_var bufnum = buf;
			if (buf_exist(ftp_job->buf)) {
				to_buffer_num(ftp_job->buf);
				if (ftp_job->operation == FTP_SEND)
					modified = 1;
			}
			if (buf_exist(ftp_job->log_buf))
				to_buffer_num(ftp_job->log_buf);
			point = size();
			refresh();
			restore_vars();
			bufnum = window_bufnum;
		}
	}
	restore_vars();
	if (num_of_popups() > 0 && !strcmp(bufname, CONNECT_LIST_BUF)) {
		int line = lines_between(0, point, 0) + 1, col = current_column();
		make_connect_list();
		go_line(line);
		move_to_column(col);
		refresh();
	}
}

do_save_all_buffers(int newfile_bufs, int *count, int *failed)
{
	int i;

	save_var bufnum, window_bufnum;
	i = buf_list(0, 0);
	do {
		bufnum = i;
		if (!*filename || !modified || buffer_not_saveable)
			continue;
		if (newfile_bufs ? (force_save_as != FSA_NEWFILE)
						 : (force_save_as == FSA_NEWFILE))
			continue;	// Only do one group.
		if (newfile_bufs) {
			if (save_all_without_asking) {
				buffer_printf(MSG_BUFFER, "save-all-buffers: Could not"
							  " save buffer %s, no file name\n", bufname);
				(*failed)++;
				continue;
			}
			to_buffer_num(i);	// Show the buffer we'll be prompting for.
		}
		switch (do_save_file(1, 1, 1)) {
			case 1:
				buffer_printf(MSG_BUFFER, "save-all-buffers: Could not save %s\n",
							  filename);
				(*failed)++;
				break;
			case 0:
				show_text(-2, 1, "save-all-buffers: Saved %s", filename);
				(*count)++;
				break;
		}
	} while (i = buf_list(1, 1));
}

#define URL_PATTERN	"|((ftp|http):[/\\][/\\][^\n\t >]+)"

default_find_linked_file(char *file)
{
	char *pat;

	if (opsys == OS_UNIX)
		pat = "[+-9A-z~]+[/.][+-9A-z]*[--9+A-z]" URL_PATTERN;
	else
		pat = "([a-zA-Z]:)?[+-9A-z~]+<.>[+-9A-z]*[--9+A-z]" URL_PATTERN;
	point -= parse_string(-1, "[^ \t\n]*");
	if (!re_search(1, pat)) {
		to_begin_line();
		if (!re_search(1, pat))
			error("Can't find any file name on this line.");
	}
	if (strchr("\"'`<", character(matchstart)))
		matchstart++;
	if (strchr("\"'`.,>", character(matchend - 1)))
		matchend--;
	grab(matchstart, matchend, file);
}

command find_linked_file()
{
	char file[FNAMELEN * 3], curdir[FNAMELEN];

	*file = 0;
	if (is_highlight_on()) {
		save_var point, mark;
		fix_region();
		if (mark - point < FNAMELEN - 1 && !lines_between(point, mark, 0))
			grab(point, mark, file);
	}
	if (!*file)
		find_linked_file_here(file);
	get_buffer_directory(curdir);
	if (!is_remote_file(file))
		absolute(file, curdir);
	if (strlen(file) > FNAMELEN - 1)
		error("File name too long.");
	quiet_set_bookmark();
	tiled_only();
	locate_window("", file);
	do_find(file, ask_line_translate());
}

// Search for a URL at or after point, return nonzero.
int find_url_here()
{
	int start = point;
	to_begin_line();
	while (re_search(1, "(www%.|(https?|ftp):[/\\][/\\])[^\n\t <>\"]+"
					 "[^\n\t <>()\"',.]"))
		if (matchend >= start)
			return 1;
	return 0;
}

command view_web_site()
{
	char url[2048], def[2048];

	*def = 0;
	html_help_check_ok();
	save_var point, mark;
	if (is_highlight_on()) {
		fix_region();
		if (mark - point < sizeof(def) - 1 && !lines_between(point, mark, 0))
			grab(point, mark, def);
	}
	if (!*def)
		if (find_url_here())
			if (character(matchstart) == 'w') {	// No http://
				strcpy(def, "http://");
				grab(matchstart, matchend, def + strlen(def));
			} else
				grab(matchstart, matchend, def);
	restore_vars();
	get_strdef(url, "View web site", def);
	if (!*url)
		return;
	if (is_win32)
		do_shell_execute(url, "Open", "view web page");
	else
		show_url(url);
}

guess_mode_without_extension(char *res, char *pat)
{
	if (!strfcmp(get_tail(filename), "makefile")
		  || !strnfcmp(get_tail(filename), "makefile.", 9)) {
		sprintf(res, pat, "mak");	// Recognize special file names
		return 1;					// w/o .mak extension.
	}
	if (mode_keys == telnet_tab) {
		sprintf(res, pat, "telnet");
		return 1;
	}		
	save_var point = 0;
	if (parse_string(1, "#%!.*/perl")) {
		sprintf(res, pat, "perl");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(%!DOCTYPE|html|xml|xst|%!--)")) {
		sprintf(res, pat, "html");
		return 1;
	}
	if (parse_string(1, "(:.*\n)?#%!.*/(sh|csh|bash|ksh)</word>|:\n")) {
		sprintf(res, pat, "sh");
		return 1;
	}
	if (parse_string(1, "%$title ")) {
		sprintf(res, pat, "gms");
		return 1;
	}
	if (parse_string(1, "%%%!PS-Adobe-")) {
		sprintf(res, pat, "ps");
		return 1;
	}
	if (size() > 1000)		// Look in 1st 1000 chars only.
		save_var narrow_end = size() - 1000;
	if (parse_string(1, "[ \t\n]*(//|/%*|#)")
		  && re_search(1, "^[ \t]*(#define|#include)[ \t]+")) {
		// It starts with a comment or preprocessor line, and it's
		// got a #define or #include, so assume it's a C++ header
		// file.
		sprintf(res, pat, "hpp");
		return 1;
	}
	point = 0;
	if (translation_type == FILETYPE_UNIX && parse_string(1, "([ \t]*\n)*#")) {
		sprintf(res, pat, "conf");
		return 1;
	}
	return 0;
}

grep_prompt(flags, inbufs)
{
	char filepat[FNAMELEN], old_cd[FNAMELEN], buf_dir[FNAMELEN];
	jmp_buf *old_level = top_level, this_level;

	_last_was_grep = 1;
	tiled_only();
	iter = 0;
	error_if_input(1);	/* error if we're in the middle of input */
	flags = get_search_string("List lines with: ",
							  flags | (inbufs ? BUFGREP : FILEGREP));
	flags &= ~(INCR | FAILING | BAD | REVERSE);
	if (!inbufs)
		save_var full_path_on_mode_line = 1;
	all_must_build_mode = 1;
	if (!get_grep_file(inbufs, filepat))
		return;
	restore_vars();
	all_must_build_mode = 1;
	top_level = &this_level;
	getcd(old_cd);
	if (grep_default_directory == 3 && get_buffer_directory(buf_dir))
		chdir(buf_dir);
	if (!setjmp(top_level))
		do_grep("grep", default_search_string(flags), flags, filepat, inbufs);
	top_level = old_level;
	if (grep_default_directory == 3)
		chdir(old_cd);
}

get_grep_file(inbufs, filepat)
char *filepat;
{
	char pr[FNAMELEN], dir[FNAMELEN];

	if (!inbufs) {
		build_prompt(pr, "In file(s)", _grep_files, 0,
					 (grep_default_directory == 3) ? 0 : 1);
		save_var prompt_with_buffer_directory = grep_prompt_with_buffer_directory;
		get_file_dir(filepat, pr);
		if (!*filepat && grep_default_directory) {
			save_var _help_on_cmd = last_index;
			if (grep_default_directory == 3)
				strcpy(filepat, _grep_files);
			else
				get_last_recall(f_match, filepat);
			if (grep_default_directory >= 2 && *filepat
				  && get_buffer_directory(dir)) {
				if (grep_default_directory != 3)
					absolute(filepat, dir);
				return 1;
			}
		}
		if (*filepat) {
			if (is_directory(filepat))
				sprintf(filepat + strlen(filepat), "%c*", path_sep);
			strcpy(_grep_files, filepat);
		}
		if (grep_default_directory == 3)
			strcpy(filepat, _grep_files);
		else
			relative(_grep_files, filepat);
		if (!*filepat)
			return 0;
	} else
		comp_read(filepat, "In buffers [pattern]: ", b_match, 0, bufname);
	return 1;
}

do_grep(buf, pat, flags, filepat, inbufs)
char *buf, *pat, *filepat;
{
	int res = 0, aborted = 0, cnt = 0, tempbuf = 0, i, skipped = 0;
	int orig, before_head, this_err;
	char *s, rel[FNAMELEN], abs[FNAMELEN], search_type[80], *(*finder)();
	spot start;

	create(buf);
	if (grep_empties_buffer)
		zap(buf);
	locate_window(buf, "");
	point = size();
	if (size())
		bprintf("\n");	/* separate from previous query */
	make_search_prompt(search_type, flags & SMODES);
	before_head = point;
	getcd(abs);
	bprintf("Found 0 lines with %s\"%s\" in %s (cur dir \"%s\")\n",
			search_type, pat, filepat, abs);
	set_character_color(before_head, point, color_class grep_header);
	start = alloc_spot(0);
	if (display_column.default != -1)
		display_column = 0;	/* horizontally scroll this window */
	grep_mode();
	finder = give_grep_finder(inbufs, filepat, &s);	// get 1st match
	save_var readonly_warning |= ROWARN_GREP, no_popup_errors = 1;
	readonly_warning &= ~(ROWARN_MSG | ROWARN_BELL);
	orig = bufnum;
	for (; s; s = (*finder)(filepat, 0)) {
		if (!inbufs && !ok_grep_file_match(s)) {
			skipped++;
			continue;
		}
		grep_make_names(rel, abs, s, inbufs);
		noteput("%d matches, searching %s...", cnt, rel);
		if (inbufs)	// skip over grep results buffer
			this_err = !strcmp(s, buf);
		else if (!look_file(abs)) {
			this_err = find_it(abs, default_translation_type);
			if (!grep_keeps_files)
				tempbuf = bufnum;
		} else {
			this_err = 0;
			if (check_dates(0) == 2)
				grep_date_warning(buf, filename);
		}
		if (this_err == EREADABORT || user_abort) {
			aborted = 1;
			break;
		} else if (this_err) {
			window_bufnum = orig;
			if (tempbuf)
				buf_delete(tempbuf);
			tempbuf = 0;
			continue;
		}
		if (grep_show_absolute_path == 2)
			convert_to_8_3_filename(abs);
		res = copy_matches(name_to_bufnum(buf), flags, pat,
						   grep_show_absolute_path ? abs : rel);
		window_bufnum = orig;
		if (res < 0)	/* error occurred during search */
			break;
		else if (res > 0) {
			cnt += res;
			build_first = 1;
			point = before_head;
			delete(point, point + parse_string(1, "Found [0-9]+ "));
			bprintf("Found %d ", cnt);
			set_character_color(before_head, point, color_class grep_header);
			point = size();
			refresh();
		}
		if (tempbuf)
			buf_delete(tempbuf);
		tempbuf = 0;
	}
	window_bufnum = orig;
	point = i = *start;
	free_spot(start);
	build_first = 1;
	if (tempbuf)
		buf_delete(tempbuf);
	say("");
	if (aborted)
		quick_abort();
	if (res == -DSBAD)
		error("Illegal search pattern: %s", pat);
	else if (res != -DSABORT) {
		locate_window(buf, "");
		modified = 0;
		point = i;
		if (skipped)
			say("Skipped %d files due to grep-ignore-file-extensions", skipped);
	}
}

prompt_file_replace(flags, inbufs)
{		// prompt for replacing in files/buffers, then do it
	char *str, with[FNAMELEN], filepat[FNAMELEN];
	char old_cd[FNAMELEN], buf_dir[FNAMELEN];
	jmp_buf *old_level = top_level, this_level;

	flags = get_replace_text(flags | (inbufs ? BUFGREP : FILEGREP),
							 with);
	if (flags == -1 || !get_grep_file(inbufs, filepat))
		return;
	str = default_search_string(flags);
	top_level = &this_level;
	getcd(old_cd);
	if (grep_default_directory == 3 && get_buffer_directory(buf_dir))
		chdir(buf_dir);
	getcd(buf_dir);
	if (!setjmp(top_level))
		do_file_replace(str, with, flags, filepat, inbufs);
	top_level = old_level;
	if (grep_default_directory == 3)
		chdir(old_cd);
}

char *man_match(s, flags)
char *s;
{
	char pat[FNAMELEN], match[FNAMELEN], *path, *t, *f, *q;
	char man_path[FNAMELEN * 2];

	if ((flags & STARTMATCH) && *s && (!isalpha(*s) || s[1] == ' ')) {
		zap(_MATCH_BUF);	// Something like "-k foo" or "2 bar".
		buffer_printf(_MATCH_BUF, "%s%s\n", s, (flags & LISTMATCH) ? "" : " ");
		// Just return text itself as the match.
	} else if (flags & STARTMATCH) {
		show_text(-1, 20, "Listing matches...");	// Show for at least .2 secs.
		zap(_MATCH_BUF);
		path = getenv("MANPATH");
		if (!path)
			build_man_path(path = man_path);
		while (!user_abort) {
			t = strchr(path, path_list_char);
			if (t)
				strncpy(pat, path, t - path);
			else
				strcpy(pat, path);
			if (*pat && pat[strlen(pat) - 1] != '/')
				strcat(pat, "/");
			sprintf(pat + strlen(pat), "man*/%s*", s);
			f = do_file_match(pat, STARTMATCH);
			for (; f && !user_abort; f = do_file_match(pat, 0)) {
				strcpy(match, get_tail(f, 0));
				q = get_extension(match);
				if (q && !strcmp(q, ".gz"))
					*q = 0;
				q = get_extension(match);
				if (!q || !*q)
					continue;
				*q = 0;
				buffer_printf(_MATCH_BUF, "%s\n", match);
			}
			if (!t)
				break;
			path = t + 1;
		}
		save_var bufname = _MATCH_BUF;
		point = 0;
		do_sort_region(0, size(), 0, 0);
		do_uniq(1, 1, 0);
		restore_vars();
	}
	return general_matcher(s, flags);
}

color_html_range(from, to) // recolor just this section
{			// last colored region may go past to
	int s, script, t, i;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	if (from < to) {
		set_character_color(from, to, color_class html_text);
		set_tagged_region("needs-color", from, to, -1);	// Say it needs coloring.
	}
	point = from;
	save_var case_fold = 1;
	color_html_move_back();
	while (point > narrow_start
		   && get_tagged_region("color-as-unit", point - 1, &from) != -1)
		point = from;		// Move before script areas.
	if (to > size() - narrow_end)
		to = size() - narrow_end;
	set_tagged_region("color-as-unit", from, to, -1);
	from = point;
	for (s = point; point < to; ) {
		if (curchar() == '&') {
			set_character_color(s, point, color_class html_text);
			s = point++;	 // Color an entity like &amp; or &#251;
			if (parse_string(1, "([a-z]+|#[0-9]+|#x[0-9a-f]+);?"))
				point = matchend;
			set_character_color(s, point, color_class html_entity);
			s = point;
			continue;
		}
		if (curchar() != '<') {
			point++;
			continue;
		}
		set_character_color(s, point, color_class html_text);
		i = t = point++;
		while (strchr(" \t\n", curchar()))
			point++;	// Parse < followed by whitespace.
		set_character_color(t, point, color_class html_punctuation);
		if (curchar() == '!')
			color_html_comment();
		else if (strchr("%?", curchar()))
			color_html_include();
		else if (s = parse_string(1, "[-_./a-z0-9]*")) { // Parse the tag.
			set_character_color(point, point + s, color_class html_tag);
			script = (s == 6 && parse_string(1, "script"));
			if (s == 3 && parse_string(1, "jsp:(declaration|expression"
									   "|scriptlet)</word>"))
				script = 2;
			point += s;			// Look for javascript tag.
			if (curchar() == ':' && parse_string(1, ":[-_./a-z0-9]*")) {
				set_character_color(point - s, point + 1,
									color_class html_namespace);
				set_character_color(point + 1, matchend,
									color_class html_tag);
				point = matchend;
			}
			for (i = point; strchr(" \t\n", curchar()); )
				point++;	// Parse whitespace.
			set_character_color(i, point, color_class html_punctuation);
			color_html_attributes(script, t);
		}
		s = point;
	}
	re_search(1, "![<&]");
	if (s < point)
		set_character_color(s, point, color_class html_text);
	set_tagged_region("needs-color", from, point, 0);	// Say it's colored.
	return point;
}

do_html_indent(int manual)
{
	int region, start;

	region = get_tagged_region("color-as-unit", point, &start);
	if (IS_ENTER_KEY(key) && !(html_auto_indent &
							   (region == 1 ? 2 : region == 2 ? 4 : 1)))
		return;
	save_var narrow_start = start;
	switch (region) {
		case 1:
			html_exclude_script_head();
			save_var c_extra_keywords |= JAVA_KEYWORDS;
			do_c_indent();
			return;
		case 2:
			html_exclude_script_head();
			do_vbasic_indent();
			return;
	}
	if (manual)
		indent_previous();
	else
		prev_indenter();
}

html_indenter()
{
	do_html_indent(0);
}

html_indent() on html_tab['\t']
{
	do_html_indent(1);
}

int html_move_level();
int html_auto_show_delimiter();
int html_recolor_from_here();

command html_mode()
{
	mode_default_settings();
	mode_keys = html_tab;		/* Use these keys. */
	major_mode = _html_mode_name;

	strcpy(comment_start, "%<%!-- *");
	strcpy(comment_pattern, "%<%!--(.|<newline>)*--%><FirstEnd>");
	strcpy(comment_begin, "<!-- ");
	strcpy(comment_end, " -->");
	indenter = html_indenter;
	auto_indent = 1;
	mode_move_level = html_move_level;
	if (auto_show_html_delimiters)
		auto_show_matching_characters = html_auto_show_delim_chars;
	mode_auto_show_delimiter = html_auto_show_delimiter;
	recolor_range = color_html_range;	// set up coloring rules
	recolor_from_here = html_recolor_from_here;
	idle_coloring_size = 5000;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	try_calling("html-mode-hook");
	drop_all_colored_regions();
	make_mode();
}

suffix_prx()
{
	html_mode();
}

suffix_asx()
{
	html_mode();
}

rectangle_standardize()
{
	int pcol, mcol;

	if (point < mark)		// First make sure point > mark.
		exchange_point_and_mark();
	pcol = virtual_column();
	mcol = virtual_mark_column();
	if (pcol < mcol) {		// Now maybe fix columns.
		to_virtual_column(mcol);
		exchange_point_and_mark();
		to_virtual_column(pcol);
		exchange_point_and_mark();
	}
}

command indent_rigidly()
{
	int indent, colstart = 0, colend;
	int orig_column, orig_mark_column;

	if (is_highlight_on() == REGRECT)
		rectangle_standardize();
	orig_column = virtual_column();
	orig_mark_column = virtual_mark_column();
	save_spot mark, point;
	if (point > mark)
		exchange_point_and_mark();
	if (!has_arg)
		iter = get_soft_tab_size();
	if (is_highlight_on() == REGRECT) {
		save_var indent_with_tabs = 0;	// Make sure point and mark stay valid.
		colstart = MIN(orig_column, orig_mark_column);
		colend = MAX(orig_column, orig_mark_column);
	} else if (is_highlight_on() == REGLINE)
		mark++;
	to_begin_line();
	while (point < mark) {
		move_to_column(colstart);
		re_search(1, "[ \t]*");
		indent = current_column() + iter;
		if (indent < colstart)
			indent = colstart;
		to_column(indent);
		if (curchar() == '\n')
			delete_horizontal_space();
		nl_forward();
	}
	restore_vars();
	if (region_type() == REGRECT) {
		if (point < mark)
			exchange_point_and_mark();
		exchange_point_and_mark();
		to_virtual_column(colstart);
		exchange_point_and_mark();
		to_virtual_column(MAX(colstart, colend + iter));
	}
	iter = 0;
}

set_mark_cmd(type)	// set mark equal to point, and begin defining
{			// the specified type of region
	int now = is_highlight_on();	// get type of current region

	if (!now || type == now && !mark_unhighlights) {
		do_set_mark(point); // start new region here
		say("Mark set.");
		if (type == REGRECT && mark_rectangle_expands)
			to_virtual_column(virtual_column() + 1);
	} else if (type == now && mark_unhighlights) {
		highlight_off();	// region is same type, remove it
		return;
	}
	set_region_type(type);	// define region of this type & show it
	highlight_on();
}

command kill_rectangle()
{				/* save, maybe replace w/ spaces */
	pick_rectangle(1, kill_rectangle_removes ? 1 : 2);
}

command yank()
{
	int pos = point, oldsize = size();

	get_clipboard();
	if (typing_deletes_highlight && is_highlight_on() == REGRECT) {
		check_num_kill_buffers();
		if (_kill_width[_num_kill] > 0)
			fix_region();
	}
	delete_if_highlighted();
	retrieve(0);
	if (_doing_input == DI_LINEINPUT && insert_default_response == 3
		  && pos == oldsize && size() < FNAMELEN) {
		undo_mainloop();		// Let undo add stuff back.
		restart_file_paths();
	}
}

drag_drop_handler() on reg_tab[WIN_DRAG_DROP]
{
	char buf[FNAMELEN], *s, dir[FNAMELEN];
	int line = -1, column = -1, j, dir_flags = directory_flags, first = 1;

	tiled_only();
	if (mouse_screen >= 0)
		to_mouse_window(mouse_x, mouse_y, mouse_screen);
	while (j = drag_drop_result(buf)){
		if (*buf == '+') {			// a +linenum parameter
			line = numtoi(buf + 1);
			s = strchr(buf, ':');
			if (s)
				column = numtoi(s + 1);
		} else if (*buf == '-') {
			switch (buf[1]) {
				case 'd':
					s = index(buf + 2, '=');
					if (!s)		// accept either VAR=VAL or VAR!VAL
						s = index(buf + 2, '!');
					if (!s)
						break;
					*s = 0;
					if (j = find_index(buf + 2)) {
						set_var_val(j, s + 1, 1);
						continue;
					} else
						say("No such variable %s.", buf + 2);
					break;
				case 'l':
					load_commands(buf + 2);
					break;
				case 'r':
					try_calling(buf + 2);
					break;
				case 'w':
					dir_flags = numtoi(buf + 2);
					break;
			}
			continue;
		} else if (*buf) {
			absolute(buf);
			if (is_directory(buf) || is_remote_file(buf) || (is_pattern(buf) & 2))
				do_find(buf, default_translation_type);
			else
				find_it(buf, default_translation_type);
			if (line >= 0) {
				go_line(line);
				line = -1;
			}
			if (column >= 0) {
				move_to_column(column);
				column = -1;
			}
			if ((dir_flags & DIR_CWD_FROM_FILE) && first) {
				strcpy(dir, buf);
				*get_tail(dir, 1) = 0;
				chdir(dir);
			}
			first = 0;
			if (j == 2)
				show_binding("Use %s to release the Epsilon client when you finish.",
							 "resume-client");
		}
	}
}

process_find_directory(char *file, int buf, int pos)
{
	char dir[FNAMELEN];
	int level = 0, got = 0, c;

	save_var bufnum = buf, point = pos;
	*dir = 0;
	if (is_process_buffer(bufnum) && *process_current_directory)
		strcpy(dir, process_current_directory);
	while (!got && re_search(-1, "(Entering|Leaving|Running `.*' in)"
							 " directory `(.*)'$")) {
		switch (c = character(find_group(1, 1))) {
			case 'L':	level++; break;
			case 'E':	if (--level > 0)
							break;
			case 'R':	got = c;
						grab(find_group(2, 1), find_group(2, 0), dir);
						break;
		}
	}
	if (got != 'R')	// "Running" message is from us, so don't translate it.
		convert_cygwin_filename(dir);
	if (*dir)
		absolute(file, dir);
	else
		absolute(file);
}

process_get_previous_cmd(int down)
{
	int buf = bufnum;
	char cmd[FNAMELEN];

	point = size();
	if (!exist(PROC_CMDS_BUF))
		return;
	*cmd = 0;
	if (has_arg) {
		iter = 0;
		save_var bufname = PROC_CMDS_BUF;
		if (point == size())
			point--;
		to_begin_line();
		restore_vars();
		if (!get_choice(name_to_bufnum(PROC_CMDS_BUF), cmd,
						"Previous commands", "Select a previous command",
						"OK", "Cancel", ""))
			return;
	}
	if (spot_to_buffer(process_input_starts_at) <= 0)
		process_input_starts_at = alloc_spot();
	if (lines_between(*process_input_starts_at, point))
		*process_input_starts_at = type_point;
	if (!lines_between(*process_input_starts_at, point))
		delete(*process_input_starts_at, point);
	else
		*process_input_starts_at = point;
	if (*cmd)
		stuff(cmd);
	else {
		save_var bufname = PROC_CMDS_BUF;
		down_or_up_lines(down);
		buf_xfer(buf, give_begin_line(), give_end_line());
	}
}

struct {
	spot start;
	spot end;

	char prefix[100];

	char found[200];
	int savepos;
	int len; 
	int pullsrc;
} pull;

command pull_word()
{
	int pl_key, old_mod = modified, orig_key = key;

	check_modify(bufnum);
	iter = 1;
	if (at_word_interior()){
		note("Can't pull word here.");
		return;
	}


	alloc_pl();

	for (pl_key = pl_trans(key); 1; pl_key = pl_getkey()){
		if (key == orig_key && pl_key != 1 && pl_key != -1)
			pl_key = 1;
		switch(pl_key){
			case 1:
			case -1:
				if (pl_search(pl_key) == 0){
					expire_message = 1;
					pl_restore();
					free_pl();
					return;
				}
				break;
			case 7:
				pl_restore();
				free_pl();
				if (*pull.found)
					user_abort = 0;
				else
					ungot_key = key;
				modified = old_mod;
				return;
			case '?':
				help_on_command("pull-word");
				break;
			default:
				free_pl();
				/* Run the key that caused us
				 * to exit.
				 */
				ungot_key = key;
				return;
		}
	}
}

command replace_again()
{	// right now, we can only replace forward
	int flags = search_flags_with_direction(UNKNOWN) | QUERY;
	char tmp[FNAMELEN], *s, *t;

	if (_buf_readonly)
		check_modify(bufnum);
	make_search_prompt(tmp, flags);
	say("%sReplace string: %s", tmp, default_search_string(flags));
	s = default_search_string(flags);
	t = (flags & REGEX) ? _default_regex_replace : _default_replace;
	last_index = find_index(help_on_search(flags));
	show_replace(s, t, flags);
	string_replace(s, t, flags | INCR);
	iter = 0;
	say("");
}

gen_replace(flags)	/* get input, then do replace */
{
	char *str, with[FNAMELEN];

	if (_buf_readonly)
		check_modify(bufnum);
	if (has_arg)
		flags |= WORD;
	save_var point;
	flags = get_replace_text(flags, with);
	restore_vars();
	str = default_search_string(flags);
	if (flags != -1) {
		show_replace(str, with, flags);
		string_replace(str, with, flags | INCR);
	}
}

string_replace(str, with, flags)
char *str, *with;
{
	spot orig, prev, oldmark;
	int skip = 0, show = 0, ret, evershow = 0, oldstart, null_detect = -1, new_null_detect;
	int perform_case_replace, this_case_replace;

	if (_buf_readonly)
		error("Buffer is read-only.");
	if (!*str)
		return 0;
	perform_case_replace = want_case_replace(str, with, flags);
	orig = alloc_spot();
	prev = alloc_spot();
	save_var postpone_asynchronous_notify = 1;
	replace_num_found = replace_num_changed = 0;
	for (;;) {
		skip = 0;
		*prev = point;
		oldstart = matchstart;
		ret = do_searching(flags, str);
		if (!ret || ret == DSBAD) {
			point = *orig;
			break;
		} else if (ret == DSABORT)
			break;
		new_null_detect = (flags & REVERSE) ? point : (size() - point);
		if (null_detect == new_null_detect) {
			if ((flags & REVERSE) ? (point-- <= 0) : (point++ >= size())) {
				point = *orig;
				break;	// Make sure replacing a nullable pattern like $
			}			// moves forward.
			continue;
		} else
			null_detect = new_null_detect;
		replace_num_found++;
ask:
		if (show) {
			show_replace(str, with, (flags & ~INCR));
			evershow = 1;
			show = 0;
		}
		if (!(flags & QUERY)) goto do_replace;
		if (!get_replace_key(flags, str, skip)) { // show match, then get key
			point = *orig;		// match went away (maybe flags changed)
			break;
		}
		if (key == '?' || key == 'h'
			  || binding_of(key, root_keys) == (short) help) {
			help_on_command("query-replace");
			show = 1;
			point = *prev;
			null_detect = -1;
			continue;
		} else if (user_abort || key == abort_key) {
			point = *orig;
			ret = DSABORT;
			goto done;
		}
		switch(tolower(key)) {
			case '^': point = *prev; matchstart = oldstart;
					  null_detect = -1;
			case '\b':
			case GREYBACK:
			case 'n': skip = 1;	/* go to the next match */
			case ' ':
			case 'y':		/* do the replace */
			case '!':
			case ',':
			case '.': break;
			case CTRL('R'):
				say("");
				null_detect = -1;
				oldmark = alloc_spot();
				*oldmark = mark;	/* save mark */
				mark = matchstart;
					/* set temporary mark before text */
				recursive_edit();
				if (spot_to_buffer(oldmark) < 0) {
					ret = DSABORT;	// user deleted buf
					goto done;
				}
				mark = *oldmark;	/* restore mark */
				free_spot(oldmark);
				show = 1;
				continue;
			default:	if (set_flag_modes(&flags)) {
						show = 1;
						goto ask;
					}
					if (!IS_ENTER_KEY(key))
						ungot_key = key;
			case ESC:	ret = 1; goto done;
		}
do_replace:
		if (!skip) {
			save_spot matchstart;
			if (perform_case_replace)
				this_case_replace = check_for_case_replace(matchstart, point);
			if (flags & REGEX)
				fromorig(with);
			else {
				delete(matchstart, point);
				stuff(with);
			}
			if (perform_case_replace)
				switch (this_case_replace) {
					case 1: do_uppercase_region(matchstart, point); break;
					case 2: do_capitalize_region(matchstart, point); break;
				}
			replace_num_changed++;
			restore_vars();
		}
		if (flags & QUERY)
			switch (key) {
			case ',': case '^':	skip = 1; goto ask;
			case '.':		ret = 1; goto done;
			case '!':		flags &= ~QUERY; show = 1;
			}
	}
done:
	if (ret == DSABORT)
		say("Replacement canceled");
	else if (ret == DSBAD)
		show_replace(str, with, (flags & ~INCR) | BAD);
	else if (!in_macro())
		if (evershow || (flags & INCR))
			say("Replaced %d of %d.", replace_num_changed,
			    replace_num_found);
		else if (evershow)
			say("");
	free_spot(orig);
	free_spot(prev);
	return ret;
}

word_search(dir, str)		/* search for whole words only */
char *str;
{
	char *pat, *s, *t;
	int ret, a = abort_searching, c;

	if (a == ABORT_JUMP)
		save_var abort_searching = ABORT_ERROR;
	pat = malloc(strlen(str) * 8 + 20);
	*pat = 0;
	if (isident(str[0]) || str[0] >= 0x80)
		strcpy(pat, "</word>");
	for (s = str, t = pat + strlen(pat); *s; s++)
		if (isalnum(*s))
			*t++ = *s;
		else if (*s == ' ') {
			strcpy(t, "<wspace>+");
			t += strlen(t);
		} else {
			*t++ = '%';
			*t++ = *s;
		}
	if (strlen(str) > 1 && isident(c = str[strlen(str) - 1]) || c >= 0x80)
		strcpy(t, "</word>");
	ret = re_search(dir, pat);
	free(pat);
	if (ret == ABORT_ERROR && a == ABORT_JUMP)
		quick_abort();
	return ret;
}

struct find {
	char *find_text;
	char *replace_text;
	int flags;
} find_data;		// The find dialog looks at this variable directly.
int our_matchstart;

perform_replace()
{
	point = our_matchstart;
	if (find_from(find_data.flags, find_data.find_text, our_matchstart,
				  our_matchstart) != 1 || our_matchstart != matchstart)
		return;		// Match has moved (flags changed?)
	if (find_data.flags & REGEX)
		fromorig(find_data.replace_text);
	else {
		delete(matchstart, point);
		stuff(find_data.replace_text);
	}
}

ev_header()
{
	if (lookpath("include/eel.h", 0))
		bprintf("#include <eel.h>\n");
	else
		bprintf("#include \"eel.h\"\n");
}

void eval_template(char *expr, char *src)
{
	char quoted[300], *p, *q;

	save_var bufnum = zap(src);
	for (p = expr, q = quoted; *q = *p; p++, q++)
		if (*p == '"') {
			*q++ = '\\';
			*q = '"';
		}
	ev_header();
	bprintf("ev_template(){\n");
	bprintf("  if (sizeof(%s) <= sizeof(int))\n", expr);
	bprintf("    say(\"%%s == %%d\", \"%s\", (%s));\n", quoted, expr);
	bprintf("  else\n");
	bprintf("    say(\"%%s == %%s\", \"%s\", (%s));\n", quoted, expr);
	bprintf("}\n");	
}

void execute_eel_template(char *expr, char *src)
{
	save_var bufnum = zap(src);
	ev_header();
	bprintf("ev_template(){ %s; }\n", expr);
}

cf_prefix()		/* handle create-prefix-command */
{
	int kt = make_anon_keytable();

	put_on_key(kt);
	set_case_indirect(index_table(kt));
	set_mouse_keys(index_table(kt));
}

cf_prefix()		/* handle create-prefix-command */
{
	int kt = make_anon_keytable();

	put_on_key(kt);
	set_case_indirect(index_table(kt));
	set_mouse_keys(index_table(kt));
}

int tag_lister();

tag_searcher_c(char *tag)
{
	char classname[FNAMELEN], fulltag[FNAMELEN], pr[FNAMELEN];
	int err, cnt = 0, selected = 0, try = 1;

	if (parse_string(-1, "-%>|%.[ \t\n]"))	// Ignore func's class.
		*classname = 0;
	else
		find_c_func_info(NULL, classname, NULL, 0);
again:
	strcpy(fulltag, tag);
	if (!strchr(tag, ':')) {		// Class not known yet.
		if (*classname && strcmp(classname, tag)) {		// See if current class 
			sprintf(fulltag, "%s::%s", classname, tag);	// has such a member.
			if (!(err = go_tag(fulltag, 1)))
				return err;
			bufnum = window_bufnum;
		}			// Is there more than one function by that name in any class?
		*fulltag = 0;
		bufnum = window_bufnum;
		if (tag_match(tag, STARTMATCH | EXACTONLY)) {
			do {
				cnt++;
			} while (tag_match(tag, EXACTONLY));
			if (cnt != 1) {
				sprintf(pr, "%d matches for %s: ", cnt, tag);
				save_var menu_width = 60, completion_lister = tag_lister;
				save_var insert_default_response = 2, tag_list_exact_only = 1;
				ungot_key = '?';
				comp_read(fulltag, pr, tag_match, (tag_case_sensitive ? 0 : COMP_FOLD)
						  | MUST_MATCH | NONE_OK | CAUTIOUS, tag);
				selected = 1;
			} else
				strcpy(fulltag, tag_match(tag, STARTMATCH | EXACTONLY));
		} else
			strcpy(fulltag, tag);
	}
	err = go_tag(fulltag, 1);
	strcpy(_last_tag, tag);
	bufnum = window_bufnum;
	if (err && !selected && can_tag_a_file(filename)) {
		// Try retagging current file before giving up.
		sprintf(pr, "Tag %s has moved, retag file %s? ",
				fulltag, filename);
		if (!tag_ask_before_retagging || try > 1
			  || ask_yn("Retag file", pr, "Yes", "No", 1)) {
			if (try < 2)
				do_retag_file(filename);
			err = go_tag(fulltag, 1);
			if (try < 2)
				do_save_tags();
			if (err && try++ < 2)
				goto again;
			if (!err)
				say("Found %s in this file.", tag);
		}
	}
	if (!err && !selected && strcmp(tag, fulltag))
		say("Searched for %s, found %s", tag, fulltag);
	report_tag_error(err);
}

go_bsc_tag(char *s)
{
	char fname[FNAMELEN];
	int linenum;

	if (!*s) {		// Old way to go to next tag.
		next_prev_tag(1);
		return 0;
	}
	if (browser_query(bsc_file, s, 0) == -1)	// error
		return 1;
	strcpy(_last_tag, s);
	if (!browser_find_match(fname, &linenum, 0))
		return 1;
	get_to_tag_match_line(fname, linenum);
	return 0;
}

can_tag_a_file(s)
char *s;
{
	char subr[FNAMELEN], def[FNAMELEN], *pattern = "tag-suffix-%s";

	s = get_extension(s);
	sprintf(subr, pattern, *s ? (s + 1) : "none");
	sprintf(def, pattern, "default");
	return !tags_from_bsc && (find_index(subr) || find_index(def));
}

good_c_tag()	// tag name at point if a valid c tag
{		// return 0 to go back to orig point before searching again
	char func[TAGLEN], tmp[TAGLEN];
	int start = point, i = 1;

	if (!parse_string(1, "[a-zA-Z0-9_]+", func))
		return 0;
	else if (!strcmp(func, "struct") || !strcmp(func, "union"))
		return tag_c_struct(0);		// maybe struct tag
	else if (!strcmp(func, "class"))
		return tag_c_struct(1);		// or similar class def
	else if (!strcmp(func, "interface") && (c_extra_keywords & JAVA_KEYWORDS))
		return tag_c_struct(1);		// for java: like a class
	else if (!strcmp(func, "enum"))
		return tag_c_struct(2);		// or similar enum def
	else if (!strcmp(func, "namespace"))
		return tag_c_struct(3);		// or namespace def
	else if (!strcmp(func, "operator") && parse_string(1, CPP_OPLIST)) {
		grab(matchstart, matchend, func + strlen(func));
		point = matchend;
	} else if (!strcmp(func, "extern")) {
		tag_extern_decl = 1;	// next vars are just declarations
		return 1;
	} else if (!strcmp(func, "__declspec")) {
		return 1;
	} else if (parse_string(-1, "^[ \t]*#[^#\n]*"))
		return tag_c_define(func);	// do preproc line
	c_fwd_space();
	if (curchar() == '_' && parse_string(1, "(_ANSI_ARGS_|__P)</word>")) {
		point = matchend;	// skip over arg-hiding macro
		c_fwd_space();
	}
	if (curchar() == '(') {
		if (matches_at(start, -1, "~[ \t]*")) {
			point = matchend;
			parse_string(1, "~[ \t]*[a-zA-Z0-9_]+", func);
		}
		i = tag_c_func(1);
	} else if (curchar() == ':' && parse_string(1, "::[ \t\n]*"
												"(_?(near|far)[ \t\n]+)?([a-zA-Z0-9_~]+|" CPP_OPSPEC ") *%(")) {
		strcat(func, "::");	// make raw class::func w/o spaces
		grab(find_group(3, 1), find_group(3, 0), func + strlen(func));
		i = tag_c_func(1);
	} else if (tag_c_variable() && ((tag_declarations & 2)
									|| !tag_extern_decl))
		i = 2;
	if ((i == 2 || i == 3) && !isdigit(*func)) {
		if (*c_tagging_class && !index(func, ':')) {
			strcpy(tmp, func);	// put class:: before func name
			sprintf(func, "%s::%s", c_tagging_class, tmp);
		}
		add_tag(func, start);
	}
	return i;
}

tag_c_func(show)
{
	int ret = 2;

	if (skip_c_delims('(', ')', 1)) {
		if (show)
			buffer_printf("-tag-errors", "Couldn't parse %s at line %d\n",
						  filename, lines_between(0, point, 0) + 1);
		return 0;	/* after identifier, skip over arg list */
	}
	c_fwd_space();
	while (tag_c_skip_before_body(show))
		c_fwd_space();
	if (parse_string(1, "on[ \t]")) {	// eel command
		search(1, "\n{");
		point--;
		ret = 2;
	} else if (parse_string(1, "[A-Za-z{]")) // func definition
		ret = skip_param_decls();
	else if (curchar() == ':')		// C++ initializer
		ret = skip_cpp_init();
	else if (curchar() == '=')		// C++ function using =0.
		return 3;
	else if (!parse_string(1, "[,;]"))	// not even a declaration
		ret = (point >= size());	// Skip parens if it looks like a macro.
	else if (!(tag_declarations & 1))	// a declaration, don't tag it?
		ret = 1;
	if (ret)
		tag_extern_decl = 0;
	return ret;
}

parse_other_class()
{
	if (curchar() == ':') {
		point++;
		for (;;) {
			c_fwd_space();
			if (!parse_string(1, "[a-zA-Z0-9_]+|,|:"))
				break;
			point = matchend;
			if (curchar() == '<')
				point += parse_string(1, "%<[a-zA-Z0-9_, \n\t]+%>");
		}
		c_fwd_space();
	} else if (is_java_buffer()) {
		while (strchr("ei", curchar())
			   && parse_string(1, "(extends|implements)</word>")) {
			point = matchend;
			skip_identifier_list();
		}
	}
	return curchar() == '{';
}

undo_it(is_undo, group_movements)
{
	int ret, before_count;
	char msg[200], *name = is_undo ? "un" : "re", *spec = 0;
	int orig_nar_start = narrow_start;
	int orig_nar_end = narrow_end;

	if (undo_count(is_undo) < 1)
		error("Nothing to %sdo.", name);
	before_count = undo_count(0);
	if (!before_count)
		_buffer_saved_here = -1;
	if (!undo_keeps_narrowing)
		narrow_start = narrow_end = 0;
	*name = tolower(*name);
	if (group_movements)
		ret = undo_group_movements(is_undo);
	else					/* plain undo */
		ret = undo_op(is_undo);
	narrow_start = orig_nar_start;
	narrow_end = orig_nar_end;
	if ((point < orig_nar_start || point > size() - orig_nar_end)
		  && !undo_keeps_narrowing) {
		say("Undoing outside narrowed region, removing narrowing.");
		narrow_start = narrow_end = 0;
	}
	*name = toupper(*name);
	if (ret & UNDO_FLAG)
		switch (undo_flag) {
			case UNDO_READ_FILE:	spec = "Buffer read from disk";
									break;
			case UNDO_WRITE_FILE:	spec = "Buffer written to disk";
									break;
			case UNDO_UNMODIFIED:	spec = "Buffer was marked unmodified";
									break;
		}
	if (spec) {
		sprintf(msg, "%s, ", spec);
		if (_buffer_saved_here == -1)
			_buffer_saved_here = undo_count(0);
		if (undo_count(0) == _buffer_saved_here || before_count == _buffer_saved_here)
			modified = 0;
	} else if (ret & (UNDO_CHANGES | UNDO_MOVE | UNDO_REDISP | UNDO_MAINLOOP))
		sprintf(msg, "%sdid %s%s%s%s%s%s%s",
				name,
				(ret & UNDO_END ? "Partial " : ""),
				(ret & UNDO_MAINLOOP ? "Command, " : ""),
				(ret & UNDO_REDISP ? "Display, " : ""),
				(ret & UNDO_MOVE ? "Move, " : ""),
				(ret & UNDO_INSERT ? "Insert, " : ""),
				(ret & UNDO_DELETE ? "Delete, " : ""),
				(ret & UNDO_REPLACE ? "Replace, " : ""));
	else
		sprintf(msg, "%sdid, ", name);
	show_text(0, 1, "%s%d bytes remain", msg, undo_count(is_undo));
	highlight_off();
	return ret;
}

when_loading()
{
	if (state_file && get_file_read_only(state_file) == 1) {
		save_var bufnum = zap(HELPBUF);
		if (strcmp(eel_version, EEL_MINOR_VERSION))
			bprintf("%s is read-only.\n"
					"Therefore it cannot be updated.\n"
					"The above state file is for Epsilon version %s.\n"
					"This Epsilon executable is from Epsilon version %s.\n"
					"Perhaps EPSPATH is set incorrectly?\n",
					state_file, eel_version, EEL_MINOR_VERSION);
		else
			bprintf("%s is read-only.\n"
					"Therefore it cannot be updated.\n"
					"But this Epsilon executable is from version %s,\n"
					"and so is the above state file.\n"
					"Therefore it appears no update is needed.\n",
					state_file, eel_version);
		view_help(" Update Not Performed ", 0);
		return;
	}

	if (strcmp("11.01", eel_version) >= 0) { // When going from 11.01 or earlier.
		cc_tab['#'] = find_index("resume_client");
		if (!dired_24_hour_time)
			dired_24_hour_time = 2;
	}

	// Load or autoload these as appropriate.
	update_autoload("perl-mode", "perl");
	update_autoload("vbasic-mode", "vbasic");
	update_autoload("python-mode", "python");
	update_autoload("hex-mode", "hex");
	update_autoload("info", "info");
	update_autoload("brief-end-key", "brief");

	strcpy(eel_version, EEL_MINOR_VERSION);
	delay(1, COND_KEY);	// Make sure the X window is up.
	refresh();
	sayput("Update complete, preparing to write state file...");
	delay(300, COND_KEY);	// Give the user a chance to abort.
	check_abort();
	if (!*state_file)
		write_state();		// Prompt for name.
	else
		do_save_state(state_file);	// Rewrite current state file.
	exit();
}
